#include<iostream>
#include<fstream>
#include<sstream>

#include <boost/filesystem.hpp>

#include <antlr/ANTLRException.hpp>

#include <HATPLexer.hpp>
#include <HATPParser.hpp>

#include <HATPPreprocessor.hh>

#if !defined USR_FUNC_EXIT_FILE
#define USR_FUNC_EXIT_FILE "userFunc.cc"
#endif

using namespace std;

//Return:
//	0: everything is fine
//	1: error with parameters
//	2: error during preprocessor phase
//	3: error during parsing
//	4: unknown error (during parsing)

int main(int argc, char *argv[])
{
	int result=0;

	cout << "************** START HATP DOMAIN EXEC **************" << endl;
	if(argc != 4){
		cerr << "ERROR: bad syntax, syntax must be HATPdomain <domain file> <costs file> <output directory>" << endl;
		result=1;
	} else {
		boost::filesystem::path costs_file_source(argv[2]);
		ifstream domain_file_checker(argv[1]), costs_file(costs_file_source.generic_string());
		std::string output_directory("");
		if(boost::filesystem::exists(argv[3])){
			if(boost::filesystem::is_directory(argv[3])){
				output_directory=argv[3];
				output_directory+="/"; //Add / to be sure it is a path. If last charactere is already / it is not a problem
			}else{
				cerr << "ERROR : " << argv[3] << " is not a directory" << endl;
				result=1;
			}
		}else{
			try{
				if(boost::filesystem::create_directory(boost::filesystem::path(std::string(argv[3])))){
					output_directory=argv[3];
				}else{
					cerr << "ERROR : can not create directory " << argv[3] << endl;
					result=1;
				}
			}catch(...){
				cerr << "ERROR : can not create directory " << argv[3] << endl;
				result=1;
			}
		}
		if(!domain_file_checker) {
			cerr << "ERROR: when opening file " << argv[1] << ", maybe this file does not exist or is not readable" << endl;
			result=1;
		} else if(!costs_file) {
			cerr << "ERROR: when opening file " << costs_file_source << ", maybe this file does not exist or is not readable" << endl;
			result=1;
		} else if(output_directory!="") { // everything OK
			Preprocessor preprocessor(argv[1]);
			if(!preprocessor.isReady()){
				cerr << "ERROR: during preprocessing step" << endl;
				result=2;
			}else{
				ifstream& domain_file=preprocessor.getCompleteDomain();
				try {
					// we read the domain description
					HATPLexer lexer(domain_file);
					HATPParser parser(lexer);
					parser.setOutput(output_directory);
					parser.hatpFile();
					// we open costs_file and we make a copy of its contents in output_directory USR_FUNC_EXIT_FILE
					stringstream costs_file_contents;
					costs_file_contents << costs_file.rdbuf();
					string usrFile = output_directory;
					usrFile += USR_FUNC_EXIT_FILE;
					ofstream outputUsrFile(usrFile.c_str(), ios::trunc); // file creation or overwriting
					outputUsrFile << "#line 1 " << absolute(costs_file_source) << endl;
					outputUsrFile << costs_file_contents.str() << endl;
					outputUsrFile.close();
					cout << "User functions for costs copied." << endl;
					cout << "SUCCESS" << endl;
				}
				catch(HATPParserException& e) {
					cerr << "An error occurred during analysis of file " << argv[1] << ":" << endl;
					std::pair<std::string, size_t> pos=preprocessor.getLine(e.getLineNumber());
					if(e.getType()==ERROR){ cerr << "ERROR "; }
					else{ cerr<< "WARNING "; }
					cerr << pos.first << ":" << pos.second << " : " << e.getMessage() << std::endl;
					result=3;
				}
				catch (antlr::ANTLRException& e) {
					std::string message=e.toString();
					boost::cmatch what;
					if(boost::regex_match(message.c_str(), what, boost::regex("line ([[:digit:]]+):([[:digit:]]+): (.*)"))){
						if(what.size()==4){
							size_t line=atoi(what[1].str().c_str());
							std::pair<std::string, size_t> pos=preprocessor.getLine(line);
							cerr << "HATP Parser exception: " << pos.first << ":" << pos.second << ":" << what[2] << " : " << what[3] << endl;
						}else{
							cerr << "HATP Parser exception: " << e.toString() << endl;
						}
					}else{
						cerr << "HATP Parser exception: " << e.toString() << endl;
					}
					result=3;
				}
				catch(...) {
					cerr << "An unidentified error occurred during analysis of files " << argv[1] << " and " << costs_file_source << endl;
					result=4;
				}
				domain_file.close();
				domain_file_checker.close();
				costs_file.close();
			}
		}
	}
	cout << "*************** END HATP DOMAIN EXEC ***************" << endl;
	return result;
}
