#include<iostream>
#include<sstream>
#include<fstream>
#include<boost/date_time/posix_time/posix_time.hpp>

#include "Plan.hh"

using namespace boost::posix_time;
using namespace std;

map<string, float> Plan::scoreWeights;
unsigned long int Plan::planCount = 1;

/**
 * prints the plan on the standard exit
 */
void Plan::printPlan() const {
	ostringstream st;
	map<string, float> efforts = projection.getEfforts();
	pair<double, double> duration = projectionTimeCanvas->totalDurationWindow();
	string separator = "######################\n";
	st << separator << "score = " << this->getScore();
	st << ", duration = [" << duration.first << ", " << duration.second << "]" << endl;
	st << separator << "\n\nTREE (depth = " << tree.getDepth() << ")\n";
	st << separator << tree.toString();
	st << separator << "\n\nPROJECTION\n";
	for(map<string, float>::const_iterator it=efforts.begin() ; it!=efforts.end() ; ++it)
		st << "charge of " << it->first << " = " << it->second << "\n";
	efforts.clear();
	st << separator << projection.toString();
	st << separator << "\n\nABSTRACTION (depth = " << abstraction.getDepth() << ")\n";
	st << separator << abstraction.toString() << separator;
	cout << st.str() << endl;
}

/**
 * make postTreatments on a rough plan: time computation and social rules
 */
bool Plan::postTreatments(LightWorldBase& wb_at_start, SocialRuleBase const& srb, float scoreMax) {
	bool result;
	float pen;
	//we compute the TimeProjection
	projectionTimeCanvas = new TimeProjection(projection, TaskProjectionOrder);
	projectionTimeCanvas->updateLinks(wb_at_start);
	projectionTimeCanvas->computeTimeWindows();
	scoreDetails[TIME_PART] = projectionTimeCanvas->totalDuration();
	result = true;
	//we get updated version of the projection
	projection = TaskList(projectionTimeCanvas->getProjection(), false);
	if(result) {
		//social rule: wasted time 
		if(srb.wastedTimeRuleDefined()) {
			pen = srb.getWastedTimeFunctions().second(projection, *projectionTimeCanvas, srb.getWastedTimeData());
			scoreDetails[WASTED_TIME_PART] = pen;
		}
		// social rule: undesirable states
		map<string, undesirableStateFunc> const& undesirableStates = srb.getUndesirableStates();
		map<string, undesirableStateFunc>::const_iterator it1;
		for(it1 = undesirableStates.begin() ; it1 != undesirableStates.end() ; ++it1) {
			pen = (it1->second)(wb_at_start, *projectionTimeCanvas);
			scoreDetails[it1->first] = pen;
		}
		// social rule: undesirable sequences
		map<string, pair<undesirableSequenceInitFunc, undesirableSequenceFunc> > const& undesirableSequencesFunctions = srb.getUndesirableSequencesFunctions();
		map<string, pair<TaskList*, vector<std::string> > > const& undesirableSequencesData = srb.getUndesirableSequencesData();
		map<string, pair<undesirableSequenceInitFunc, undesirableSequenceFunc> >::const_iterator it2_1;
		map<string, pair<TaskList*, vector<std::string> > >::const_iterator it2_2;
		for(it2_1 = undesirableSequencesFunctions.begin() ; it2_1 != undesirableSequencesFunctions.end() ; ++it2_1) {
			it2_2 = undesirableSequencesData.find(it2_1->first);
			pen = (it2_1->second).second(*((it2_2->second).first), (it2_2->second).second, wb_at_start, projection, *projectionTimeCanvas);
			scoreDetails[it2_1->first] = pen;
		}
		//social rule: effort balancing
		if(srb.effortBalancingRuleDefined()) { 
			effortBalancingFunc fct = srb.getEffortBalancingRule();
			if(fct) { 
				pen = fct(wb_at_start, projection.getEfforts());
				scoreDetails[EFFORT_BALANCING_PART] = pen;
			}
		}
		//social rule: control of intricacy
		if(srb.intricacyControlRuleDefined()) {
			pen = srb.getIntricacyControlFunctions().second(projection, srb.getIntricacyControlData());
			scoreDetails[INTRICACY_CONTROL_PART] = pen;
		}
	}
	return result;
}

/**
 * compute contribution using scoreDetails and detailsReference
 */
float Plan::determineContribution(map<string, float> const& detailsReference) const {
	std::map<std::string, float>::const_iterator it;
	float cri, score_ref, score_plan, weight, cr=0.;
	for(it = detailsReference.begin() ; it != detailsReference.end() ; ++it) {
		score_ref = it->second;
		score_plan = scoreDetails.find(it->first)->second;
		weight = scoreWeights.find(it->first)->second;
		if((score_ref == 0) && (score_plan == 0)) cri = 0;
		else cri = ((score_ref - score_plan) / ((score_ref > score_plan)?score_ref:score_plan));
		cr += (cri * weight);
	}
	return cr;
}

/**
 * destructor: free memory by deleting tree, projection and abstraction
 */
Plan::~Plan() {
	delete projectionTimeCanvas;
}
